#!/bin/bash

#----------------------------------------------------------------------------------------------------------
#                                            Argument Parsing
#----------------------------------------------------------------------------------------------------------
SHOW_HELP=NO
PYTHON_VERSION=python3
PYTHON_BIN=undefined

echo "-----------------------------------------------------------------------------------------------------------"
echo "             Impact Acquire script for swig based Python wrapper generation for $PYTHON_VERSION            "
echo "-----------------------------------------------------------------------------------------------------------"
echo 'To compile the wrapper code it is necessary to have the python-dev package installed on this system.' 
if [[ $# -gt 0 ]] ; then
    if [ "$1" == "-h" ] || [ "$1" == "--help" ] ; then
        SHOW_HELP=YES
	elif [ "$1" == "python2" ] || [ "$1" == "Python2" ] || [ "$1" == "PYTHON2" ]; then
		echo "$1 is not supported anymore. The script will be terminated. Please compile the API for Python 3 or use Impact Acquire < 3.0.1 instead."
		exit 5
    else
        PYTHON_VERSION=$1
    fi
fi
#----------------------------------------------------------------------------------------------------------
#                                            Help & Checks
#----------------------------------------------------------------------------------------------------------
if [ "$SHOW_HELP" == "YES" ] ; then
  echo
  echo "Impact Acquire script for swig based Python wrapper generation for $PYTHON_VERSION"
  echo
  echo "Usage example (create and install wrapper code using default python 3 interpreter):       ./compileWrapperCode"
  echo "Usage Example (create and install wrapper code using default python 3 interpreter):       ./compileWrapperCode python3"
  echo "Usage Example (create and install wrapper code using python 3.8 interpreter):             ./compileWrapperCode python3.8"
  echo
  echo "Arguments:"
  echo
  echo "-h --help                       Display this help."
  echo
  echo "python3.x                       Compile the wrapper code using Python 3.x interpreter (x: the minor version number of the Python interpreter)"
  echo
  exit 22
fi

if [ "$(which "$PYTHON_VERSION")x" != "x" ]; then
    export PYTHON_BIN=$(which "$PYTHON_VERSION")
else
    echo "Python executable could not be located!!!"
    exit 42
fi

echo "-----------------------------------------------------------------------------------------------------------"
echo "                             Building Python wrapper for $PYTHON_VERSION...                                "
echo "-----------------------------------------------------------------------------------------------------------"
# check if the python module 'build' has been installed already, if not, try to install the module first
"$PYTHON_BIN" -c "import build; import setuptools" >/dev/null 2>&1
if [ $? -ne 0 ]; then
	echo
	echo "Warning: the modules 'build' and 'setuptools' are needed but at least one is not installed. Will try to install these modules via 'pip'."
	echo
	"$PYTHON_BIN" -m pip install build setuptools
	"$PYTHON_BIN" -c "import build; import setuptools"
	if [ $? -ne 0 ]; then
		echo
		echo "Error: failed to install modules 'build' and 'setuptools' via 'pip'. Please install these modules manually!"
		echo
		exit 0
	fi
fi
# nice due to linking stage taking forever on x86_64 systems
sudo -E nice -n -8 "$PYTHON_BIN" -m build

echo "-----------------------------------------------------------------------------------------------------------"
echo "                             Installing Python wrapper for $PYTHON_VERSION...                              "
echo "-----------------------------------------------------------------------------------------------------------"
sudo -E "$PYTHON_BIN" -m pip install --upgrade --no-cache-dir .

# option to change the wrapper's file permissions for non-root users in case of a high umask
if [ -z "$MVIMPACT_ACQUIRE_SOURCE_DIR" ]; then
    umask=$(umask | tail -2 | sed -n "/[0-9]/p")
    if [[ $umask -gt 002 ]] && [[ $umask -ne 022 ]] && [[ $umask -ne 020 ]]; then
        YES_NO=
        echo "-----------------------------------------------------------------------------------------------------------"
        echo "WARNING:"
        echo "Current umask $umask is detected."
        echo "The currently built Python wrapper might not be accessible for non-root users!"
        echo "-----------------------------------------------------------------------------------------------------------"
        echo "Do you want the script to change file permissions for the built Python wrapper, so that non-root users can access the wrapper as well?"
        echo "Hit 'n' + <Enter> for 'no', or just <Enter> for 'yes'."
        read -r YES_NO
        if [ "$YES_NO" == "n" ] || [ "$YES_NO" == "N" ]; then
            echo "Will NOT change file permissions for the built Python wrapper. The non-root users might not be able to access the current Python wrapper afterwards."
        else
            echo "Will change file permissions for the currently built Python wrapper now..."
            cd $($PYTHON_BIN -c "import mvIMPACT; print(mvIMPACT.__path__)" | awk -F"[']" '{print $2}')/.. || exit 1
            sudo chmod 745 mvIMPACT
            cd mvIMPACT || exit 1
            sudo chmod -R 644 "$(ls | awk '{if($1 != "Common"){print $1}}')"
            sudo chmod a+rx Common
            sudo chmod -R 644 Common/*
            echo "Done changing file permissions for the current Python wrapper."
        fi
    fi
fi

echo "-----------------------------------------------------------------------------------------------------------"
echo "Important! You may need to uninstall old versions of 'mvIMPACT Acquire' manually, since the package is now called Impact Acquire!"
echo "-----------------------------------------------------------------------------------------------------------"
