# Builds the extensions with separate SWIG run-time support
# (see http://www.swig.org/Doc1.3/Modules.html) so as to
# share datatypes among different extensions.

import platform, os, sys
from pathlib import Path, PurePath
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext

if '--help' in str(sys.argv):
    print("""
WARNING! It is no longer supported, neither by Impact Acquire nor by Python itself, to run setup.py directly.
Please use `python -m build` and `python -m pip .` for building and installing the Python wrapper instead.
Not specifying any of the options below is not guaranteed to work!

Impact Acquire specific options (always as first argument):
  return_acquire_version

Example:
  python setup.py return_acquire_version
will print something like:
  2.49.0
""")
    sys.exit()

if "MVIMPACT_ACQUIRE_SOURCE_DIR" in os.environ:
    mvimpact_acquire_dir = Path(os.environ['MVIMPACT_ACQUIRE_SOURCE_DIR'])
elif "MVIMPACT_ACQUIRE_DIR" in os.environ:
    mvimpact_acquire_dir = Path(os.environ['MVIMPACT_ACQUIRE_DIR'])
else:
    print("Either 'MVIMPACT_ACQUIRE_DIR' or 'MVIMPACT_ACQUIRE_SOURCE_DIR' must be set!")
    sys.exit(-1)

systemPlatform = platform.system()
if systemPlatform != "Linux" and systemPlatform != "Windows":
    print("Could not determine system platform")
    sys.exit(-1)

if "BUILDDIR_DN" in os.environ:
    builddir_dn = Path(os.environ['BUILDDIR_DN'])
else:
    builddir_dn = mvimpact_acquire_dir / r"LanguageBindings/Python/Output"
    if systemPlatform == "Linux":
        os.system("export BUILDDIR_DN={}".format(builddir_dn))
    elif systemPlatform == "Windows":
        os.system("set BUILDDIR_DN={}".format(builddir_dn))
version_file = mvimpact_acquire_dir / r"DriverBase/Include/mvVersionInfo.h"

class AcquireVersion(object):
    def __init__(self):
        class BuildNr(str):
            def returnNr(self, key): # key is a str like 'MVIMPACT_ACQUIRE_MAJOR_VERSION'
                offset = self.find(key)
                if offset == -1:
                    return '?'
                offset += len(key)
                while not self[offset].isdigit():
                    offset += 1
                nr = self[offset]
                while True:
                    offset += 1
                    if not self[offset].isdigit():
                        return nr
                    nr += self[offset]

        version_all_h = open(version_file)
        version = BuildNr(version_all_h.read())
        version_all_h.close()
        self.v_major = version.returnNr('#define MVIMPACT_ACQUIRE_MAJOR_VERSION ')
        self.v_minor = version.returnNr('#define MVIMPACT_ACQUIRE_MINOR_VERSION ')
        self.v_point = version.returnNr('#define MVIMPACT_ACQUIRE_RELEASE_VERSION ')
        self.v_build = version.returnNr('#define MVIMPACT_ACQUIRE_BUILD_VERSION ')

    def full(self):
        return self.v_major + '.' + self.v_minor + '.' + self.v_point + '.' + self.v_build

acquire_version = AcquireVersion()

if len(sys.argv) > 1:
    if sys.argv[1] == 'return_acquire_version':
        print('%s.%s.%s' % (acquire_version.v_major,
                            acquire_version.v_minor,
                            acquire_version.v_point))
        sys.exit()

if systemPlatform == "Linux":
    if( platform.architecture()[0] == "32bit" ):
        bits = 32
    elif( platform.architecture()[0] == "64bit" ):
        bits = 64
    else:
        print("Error! Could not determine the systems CPU architecture!")
        sys.exit(-1)
    machine = platform.machine()
elif systemPlatform == "Windows":
    bits = 64 if os.environ["PROCESSOR_ARCHITECTURE"] == "AMD64" else 32

macros = []
defines = os.getenv('DEFINES')
if defines is not None:
    for dfn in defines.split():
        macros += [(dfn[2:],'1')]

class dirlist(list):
    """You can define multiple directories in one environment variable"""
    def addDir(self, d):
        if type(d) is list:
            for a in d:
                self.addDir(Path(a)) # recurse
        else:
            d = Path(d)
            if os.pathsep in str(d): # turn 'bar;foo' into ['bar', 'foo']
                self.addDir(str(d).split(os.pathsep)) # recurse
            else:
                d = Path(str(d).replace('\ ', ' '))
                if d.is_dir():
                    self.append(d)
    def getList(self):
        return list(self)

inc_dirs = dirlist()
lib_dirs = dirlist()

# inc_dirs     : additional include directories
# lib_dirs     : additional library directories
# macros       : defined while building the extension

link_args = []
compile_args = []
# for debugging the Python extension library with Visual Studio on Windows set the variable in the next line to 'True'!
generateDebugInfo = False

installation_path = os.getenv('BUILD_FROM_INSTALLATION') # None if not set
if installation_path == '0': # '0' is explicit for 'not from installation'
    installation_path = None #  (like not specifying BUILD_FROM_INSTALLATION at all)
python_ver = '.'.join(str(x) for x in sys.version_info[:2])
if systemPlatform == "Linux":
    if "MVIMPACT_ACQUIRE_SOURCE_DIR" in os.environ:
        if ("MVIMPACT_ACQUIRE_BUILD_DIR" in os.environ):
            lib_dirs.addDir(Path(os.environ["MVIMPACT_ACQUIRE_BUILD_DIR"]) / Path('lib'))
        else:
            print()
            print("ERROR: MVIMPACT_ACQUIRE_SOURCE_DIR will be used but MVIMPACT_ACQUIRE_BUILD_DIR is not set! The build directory thus cannot be found!")
            sys.exit(-1)
    else:
        if ((machine == "armv7l") or (machine == "armv7")):
            target_lib_sub_dir = Path('lib/armhf')
        elif (machine == "i686" or machine == "x86_64") :
            target_lib_sub_dir = Path('lib/x86_64')
        elif ( machine == "aarch64") :
            target_lib_sub_dir = Path('lib/arm64')
        else:
            print()
            print("ERROR: Platform unsupported!")
            sys.exit(-1)
        lib_dirs.addDir(mvimpact_acquire_dir / target_lib_sub_dir)
    compile_args = ['-Wno-unknown-pragmas', '-Wno-misleading-indentation', '-DCOMPILE_PYTHON']
    if ((machine == "armv7l") or (machine == "armv7")):
        compile_args += ['-O0']
        link_args = ['-O0']
elif systemPlatform == "Windows":
    compile_args = ['/GR','/bigobj', '/EHsc', '/DCOMPILE_PYTHON']
    if generateDebugInfo:
        compile_args += ['/Zi']
        link_args += ['/DEBUG']
    if "MVIMPACT_ACQUIRE_SOURCE_DIR" in os.environ:
        if bits == 32:
            if ("MVIMPACT_ACQUIRE_BUILD_X86_DIR" in os.environ) and ("MVIMPACT_ACQUIRE_BUILD_CONFIGURATION" in os.environ):
                build_lib_dir = Path(os.environ["MVIMPACT_ACQUIRE_BUILD_X86_DIR"]) / r"lib\win\win32" / os.environ["MVIMPACT_ACQUIRE_BUILD_CONFIGURATION"]
            else:
                print()
                print("ERROR: MVIMPACT_ACQUIRE_SOURCE_DIR will be used but MVIMPACT_ACQUIRE_BUILD_X86_DIR is not set! The build directory thus cannot be found!")
                sys.exit(-1)
        else:
            if ("MVIMPACT_ACQUIRE_BUILD_X64_DIR" in os.environ) and ("MVIMPACT_ACQUIRE_BUILD_CONFIGURATION" in os.environ):
                build_lib_dir = Path(os.environ["MVIMPACT_ACQUIRE_BUILD_X64_DIR"]) / r"lib\win\x64" / os.environ["MVIMPACT_ACQUIRE_BUILD_CONFIGURATION"]
            else:
                print()
                print("ERROR: MVIMPACT_ACQUIRE_SOURCE_DIR will be used but MVIMPACT_ACQUIRE_BUILD_X64_DIR is not set! The build directory thus cannot be found!")
                sys.exit(-1)
        lib_dirs.addDir(build_lib_dir)
        link_args.append('/PDB:{}'.format(build_lib_dir / 'lib_mvIMPACT_acquire.pdb'))
    else:
        if bits == 32:
            target_lib_sub_dir = Path('lib')
        else:
            target_lib_sub_dir = Path(r'lib\win\x64')
        lib_dirs.addDir(mvimpact_acquire_dir / target_lib_sub_dir)

inc_dirs.addDir(mvimpact_acquire_dir)
inc_dirs.addDir(mvimpact_acquire_dir / 'mvIMPACT_CPP')
if systemPlatform == "Windows":
    inc_dirs.addDir(mvimpact_acquire_dir / 'mvDisplay' / 'Include')

build_options = {'build_base' : builddir_dn}
bdist_options = {'bdist_base' : builddir_dn, 'dist_dir' : builddir_dn}

# Create an 'Extension' object that describes how to build the Python extension.
def createExtension():
    libs = ['mvDeviceManager']

    incl_dirs = inc_dirs.getList()
    libr_dirs = lib_dirs.getList()

    global installation_path
    if installation_path is not None:
        installation_path = Path(installation_path)
        incl_dirs.insert(0, installation_path)
        incl_dirs.insert(0, installation_path / 'mvDisplay' / 'Include')
        incl_dirs.insert(0, installation_path / 'mvIMPACT_CPP')
        libr_dirs.insert(0, installation_path / 'lib')
    else:
        incl_dirs.insert(0, mvimpact_acquire_dir)
    sources_fpns = [Path(builddir_dn) / 'acquire_wrap.cpp']
    return Extension('lib_mvIMPACT_acquire',
        sources       = list(map(str, sources_fpns)),
        include_dirs  = list(map(str, incl_dirs)),
        library_dirs  = list(map(str, libr_dirs)),
        define_macros = macros,
        libraries     = libs,
        extra_compile_args = compile_args,
        extra_link_args = link_args)

class build_ext_posix(build_ext):
    # on Linux, once the lib has been built, dump a map, then strip the lib
    def build_extensions(self):
        self.check_extensions_list(self.extensions)
        for ext in self.extensions:
            self.build_extension(ext)
            libName = str(Path(self.build_lib) / 'mvIMPACT' / ext.name)
  
def getPlatform():
    platformList = []
    if systemPlatform == "Linux":
        platformList.insert(0, os.system('ldconfig -p | grep mvDeviceManager | cut -d \> -f 2 | grep -o "/lib/.*/" | cut -d/ -f 3'))
    elif systemPlatform == "Windows":
        platformList.insert(0, os.environ["PROCESSOR_ARCHITECTURE"])
    return platformList

setup(
    name = "ImpactAcquire", # remove this once we require setuptools 61.0.0 or above
    version     = acquire_version.full(),
    platforms   = getPlatform(),
    description = 'Impact Acquire Python {} interface'.format(python_ver),
    ext_package = 'mvIMPACT', # all modules are in 'mvIMPACT'
    ext_modules = [createExtension()],
    package_dir = {'mvIMPACT' : 'Output'},
    packages    = ['mvIMPACT', 'mvIMPACT.Common'],
    cmdclass    = {'build_ext' : build_ext_posix if os.name == 'posix' else build_ext},
    options     = {'build' : build_options, 'bdist' : bdist_options}
)

