//
// @description: Example applications for Impact Acquire
// @copyright: Copyright (C) 2006 - 2018 Balluff GmbH
// @authors: APIs and drivers development team at Balluff GmbH
// @initial date: 2006-06-23
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice (including the next paragraph) shall be included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,i
// WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//

//-----------------------------------------------------------------------------
#include <algorithm>
#include <apps/Common/wxAbstraction.h>
#include <common/STLHelper.h>
#include "DebugFileParser.h"
#include <functional>
#include <map>

using namespace std;

typedef map<string, string> StringMap;

const string LogConfigurationFileParser::m_ATTR_NAME( "name" );
const string LogConfigurationFileParser::m_ATTR_FLAGS( "flags" );
const string LogConfigurationFileParser::m_ATTR_OUTPUT_MASK( "outputmask" );
const string LogConfigurationFileParser::m_ATTR_OUTPUT_FILE_NAME( "outputfile" );
const string LogConfigurationFileParser::m_ATTR_OUTPUT_FILE_FORMAT( "fileFormat" );
const string LogConfigurationFileParser::m_ATTR_CLEAR_FILE( "clearFile" );
const string LogConfigurationFileParser::m_ATTR_BACKUP_PREVIOUS_FILE( "backupPreviousFile" );
const string LogConfigurationFileParser::m_ATTR_TIMESTAMP_FORMAT( "timestampFormat" );
const string LogConfigurationFileParser::m_ATTR_STYLESHEET_NAME( "stylesheet" );

//-----------------------------------------------------------------------------
LogConfigurationFileParser::TTagType LogConfigurationFileParser::GetTagType( const char* tagName ) const
//-----------------------------------------------------------------------------
{
    if( strcmp( tagName, "DebugWriter" ) == 0 )
    {
        return ttDBWriter;
    }
    else if( strcmp( tagName, "DebugWriterList" ) == 0 )
    {
        return ttDBWriterList;
    }
    return ttInvalid;
}

//-----------------------------------------------------------------------------
void LogConfigurationFileParser::OnPostCreate( void )
//-----------------------------------------------------------------------------
{
    EnableStartElementHandler();
}

//-----------------------------------------------------------------------------
unsigned int StringToBitmask( const string& input )
//-----------------------------------------------------------------------------
{
    unsigned int result = 0;
    size_t length = input.length();
    for( size_t j = 0; j < length; j++ )
    {
        if( input[length - 1 - j] == '1' )
        {
            result |= 1 << j;
        }
    }
    return result;
}

//-----------------------------------------------------------------------------
void LogConfigurationFileParser::OnStartElement( const XML_Char* pszName, const XML_Char** papszAttrs )
//-----------------------------------------------------------------------------
{
    switch( GetTagType( pszName ) )
    {
    case ttDBWriter:
        {
            // build attribute map
            StringMap attrMap;
            int i = 0;
            while( papszAttrs[i] )
            {
                attrMap.insert( pair<string, string>( papszAttrs[i], papszAttrs[i + 1] ) );
                i += 2;
            }

            LogOutputConfiguration data;
            StringMap::iterator it = attrMap.find( m_ATTR_NAME );
            if( it != attrMap.end() )
            {
                data.sectionName = it->second;
            }

            it = attrMap.find( m_ATTR_FLAGS );
            if( it != attrMap.end() )
            {
                string::size_type firstOne = it->second.find_first_of( "1" );
                // remove leading zeros...
                if( firstOne != string::npos )
                {
                    data.outputFlagMask = it->second.substr( firstOne );
                }
                else
                {
                    data.outputFlagMask = it->second;
                }
            }

            it = attrMap.find( m_ATTR_OUTPUT_MASK );
            if( it != attrMap.end() )
            {
                data.outputDestinationMask = TDebugOutputDestination( StringToBitmask( it->second ) );
            }

            it = attrMap.find( m_ATTR_OUTPUT_FILE_NAME );
            if( it != attrMap.end() )
            {
                data.outputFileName = it->second;
            }

            it = attrMap.find( m_ATTR_OUTPUT_FILE_FORMAT );
            if( it != attrMap.end() )
            {
                makeLowerCase( it->second );
                data.outputFileFormat = FileFormatFromString( ConvertedString( it->second ) );
            }

            it = attrMap.find( m_ATTR_CLEAR_FILE );
            if( it != attrMap.end() )
            {
                data.boClearFile = ( atoi( it->second.c_str() ) != 0 );
            }

            it = attrMap.find( m_ATTR_BACKUP_PREVIOUS_FILE );
            if( it != attrMap.end() )
            {
                data.boBackupPreviousFile = ( atoi( it->second.c_str() ) != 0 );
            }

            it = attrMap.find( m_ATTR_TIMESTAMP_FORMAT );
            if( it != attrMap.end() )
            {
                makeLowerCase( it->second );
                data.timestampFormat = TimestampFormatFromString( ConvertedString( it->second ) );
            }

            it = attrMap.find( m_ATTR_STYLESHEET_NAME );
            if( it != attrMap.end() )
            {
                data.stylesheetName = it->second;
            }
            m_results.push_back( data );
        }
        break;
    case ttDBWriterList:
        // nothing to be done here...
        break;
    default:
        // invalid tag...
        break;
    }
}

//-----------------------------------------------------------------------------
wxString FileFormatToString( TLogFileFormat fileFormat )
//-----------------------------------------------------------------------------
{
    switch( fileFormat )
    {
    case lffXML:
        break;
    case lffText:
        return wxString( wxT( "text" ) );
    case lffMVLOG:
        return wxString( wxT( "mvlog" ) );
    }
    return wxString( wxT( "xml" ) );
}

//-----------------------------------------------------------------------------
TLogFileFormat FileFormatFromString( const wxString& fileFormat )
//-----------------------------------------------------------------------------
{
    if( fileFormat == wxString( wxT( "text" ) ) )
    {
        return lffText;
    }
    else if( fileFormat == wxString( wxT( "mvlog" ) ) )
    {
        return lffMVLOG;
    }
    return lffXML;
}

//-----------------------------------------------------------------------------
wxString TimestampFormatToString( TTimestampFormat timestampFormat )
//-----------------------------------------------------------------------------
{
    switch( timestampFormat )
    {
    case tfLocal_ms:
        break;
    case tfRelative_ms:
        return wxString( wxT( "relative_ms" ) );
    }
    return wxString( wxT( "local_ms" ) );
}

//-----------------------------------------------------------------------------
TTimestampFormat TimestampFormatFromString( const wxString& timestampFormat )
//-----------------------------------------------------------------------------
{
    if( timestampFormat == wxString( wxT( "relative_ms" ) ) )
    {
        return tfRelative_ms;
    }
    return tfLocal_ms;
}