//
// @description: Example applications for Impact Acquire
// @copyright: Copyright (C) 2009 - 2021 Balluff GmbH
// @authors: APIs and drivers development team at Balluff GmbH
// @initial date: 2009-12-08
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice (including the next paragraph) shall be included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,i
// WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//

//-----------------------------------------------------------------------------
#include <algorithm>
#include <apps/Common/wxAbstraction.h>
#include "PackageDescriptionParser.h"
#include <map>

typedef std::map<std::string, std::string> StringMap;

using namespace std;

//-----------------------------------------------------------------------------
bool ExtractStringAttribute( const StringMap& attrMap, const std::string& attrName, wxString& destination )
//-----------------------------------------------------------------------------
{
    StringMap::const_iterator it = attrMap.find( attrName );
    if( it != attrMap.end() )
    {
        destination = ConvertedString( it->second );
    }
    return ( ( it == attrMap.end() ) ? false : true );
}

//-----------------------------------------------------------------------------
bool ExtractVersionInformation( const StringMap& attrMap, const std::string& tagName, Version& versionEntry )
//-----------------------------------------------------------------------------
{
    StringMap::const_iterator it = attrMap.find( tagName );
    if( it != attrMap.end() )
    {
        versionEntry = VersionFromString( ConvertedString( it->second ) );
    }
    return ( ( it == attrMap.end() ) ? false : true );
}

const string PackageDescriptionFileParser::m_ATTR_NAME( "Name" );
const string PackageDescriptionFileParser::m_ATTR_TYPE( "Type" );
const string PackageDescriptionFileParser::m_ATTR_DESCRIPTION( "Description" );
const string PackageDescriptionFileParser::m_ATTR_VERSION( "Version" );
const string PackageDescriptionFileParser::m_ATTR_REVISION_MIN( "RevisionMin" );
const string PackageDescriptionFileParser::m_ATTR_REVISION_MAX( "RevisionMax" );

//-----------------------------------------------------------------------------
PackageDescriptionFileParser::TTagType PackageDescriptionFileParser::GetTagType( const char* tagName ) const
//-----------------------------------------------------------------------------
{
    if( strcmp( tagName, "PackageDescription" ) == 0 )
    {
        return ttPackageDescription;
    }
    else if( strcmp( tagName, "File" ) == 0 )
    {
        return ttFile;
    }
    else if( strcmp( tagName, "SuitableProductKey" ) == 0 )
    {
        return ttSuitableProductKey;
    }
    return ttInvalid;
}

//-----------------------------------------------------------------------------
void PackageDescriptionFileParser::OnPostCreate( void )
//-----------------------------------------------------------------------------
{
    m_packageDescriptionVersion.major_ = 1;
    m_packageDescriptionVersion.minor_ = 0;
    EnableStartElementHandler();
}

//-----------------------------------------------------------------------------
void PackageDescriptionFileParser::OnStartElement( const XML_Char* pszName, const XML_Char** papszAttrs )
//-----------------------------------------------------------------------------
{
    if( ( m_packageDescriptionVersion.major_ != 1 ) ||
        ( m_packageDescriptionVersion.minor_ != 0 ) )
    {
        m_lastError = wxString::Format( wxT( "Unsupported package description version: %ld.%ld\n" ), m_packageDescriptionVersion.major_, m_packageDescriptionVersion.minor_ );
        return;
    }

    // build attribute map
    StringMap attrMap;
    int i = 0;
    while( papszAttrs[i] )
    {
        attrMap.insert( pair<string, string>( papszAttrs[i], papszAttrs[i + 1] ) );
        i += 2;
    }

    switch( GetTagType( pszName ) )
    {
    case ttPackageDescription:
        {
            if( !ExtractVersionInformation( attrMap, m_ATTR_VERSION, m_packageDescriptionVersion ) )
            {
                m_lastError = wxString( wxT( "Version information tag is missing in package format description" ) );
                m_packageDescriptionVersion = Version();
                break;
            }
        }
        break;
    case ttFile:
        {
            FileEntry entry;
            if( !ExtractStringAttribute( attrMap, m_ATTR_DESCRIPTION, entry.description_ ) ||
                !ExtractStringAttribute( attrMap, m_ATTR_NAME, entry.name_ ) ||
                !ExtractStringAttribute( attrMap, m_ATTR_TYPE, entry.type_ ) ||
                !ExtractVersionInformation( attrMap, m_ATTR_VERSION, entry.version_ ) )
            {
                m_lastError = wxString( wxT( "At least one required attribute for the file tag is missing" ) );
                break;
            }
            entry.packageDescriptionVersion_ = m_packageDescriptionVersion;
            m_results.push_back( entry );
        }
        break;
    case ttSuitableProductKey:
        {
            if( m_results.empty() )
            {
                m_lastError = wxString( wxT( "Orphaned 'SuitableProductKey' tag detected" ) );
                break;
            }

            SuitableProductKey data;
            if( !ExtractStringAttribute( attrMap, m_ATTR_NAME, data.name_ ) ||
                !ExtractVersionInformation( attrMap, m_ATTR_REVISION_MAX, data.revisionMax_ ) ||
                !ExtractVersionInformation( attrMap, m_ATTR_REVISION_MIN, data.revisionMin_ ) )
            {
                // at least one required attribute is missing
                break;
            }

            m_results[m_results.size() - 1].suitableProductKeys_.insert( make_pair( data.name_, data ) );
        }
        break;
    default:
        // invalid tag...
        break;
    }
}
