#ifndef __MVPCI_H__
#define __MVPCI_H__

#ifdef __KERNEL__

/*
 * mvPCIe DMA Driver for BVS CA-BN (mvBlueNAOS) series cameras
 *
 * SPDX-FileCopyrightText: 2019 - 2023 MATRIX VISION GmbH <info@matrix-vision.de>
 * SPDX-FileCopyrightText: 2023 Balluff GmbH
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */

#include <linux/interrupt.h>
#include <linux/device.h>
#include <linux/cdev.h>
#include <linux/mutex.h>
#include <linux/spinlock.h>
#include <linux/uio_driver.h>
#include <linux/pci.h>
#include <linux/i2c.h>
#include <linux/hrtimer.h>

#include "mvdma_ioctl.h"

#define MVPCI_DRIVER_NAME "mvpci"

#define MVPCI_MAJOR_VERSION 0
#define MVPCI_MINOR_VERSION 5
#define MVPCI_MINI_VERSION  0 
#define MVPCI_VERSION                                                   \
    (MVPCI_MAJOR_VERSION << 8) | (MVPCI_MINOR_VERSION << 4)             \
        | MVPCI_MINI_VERSION
#define MVPCI_VERSION_STR                                               \
    __stringify(MVPCI_MAJOR_VERSION) "."__stringify(                    \
        MVPCI_MINOR_VERSION) "."__stringify(MVPCI_MINI_VERSION)

#define ISR_DEBUG               /* without this you will get absolutely no debug messages in \
                                   the isr */

#define MVPCI_FLAG_RES0 0x01
#define MVPCI_FLAG_RES1 0x02
#define MVPCI_FLAG_RES2 0x04
#define MVPCI_FLAG_RES3 0x08
#define MVPCI_FLAG_IRQ  0x10
#define MVPCI_FLAG_MSI  0x20
#define MVPCI_FLAG_IORW 0x40

#define dprintk(lvl, msg...)                                            \
    do {                                                                \
        if (debug >= lvl)                                               \
            printk("mvpci: " msg);                                      \
    } while (0)
#ifdef ISR_DEBUG
#define iprintk(lvl, msg...)                                            \
    do {                                                                \
        if (isr_debug >= lvl)                                           \
            printk("mvpci-isr: " msg);                                  \
    } while (0)
#else
#define iprintk(lvl, msg...)                                            \
    do {                                                                \
    } while (0);
#endif

/*********************** structs + macros ******************/

//#define DEBUG_MVWRITE
#ifdef DEBUG_MVWRITE
#define mvwrite(dat, adr)                                                   \
    {                                                                       \
        dprintk(0,                                                          \
            "mvpci-%d: writing 0x%08x to register @ 0x%08x (0x%08lx)...\n", \
            mvpci_drv->nr, (dat), (adr), mvpci_drv->mmio + (adr));          \
        writel((dat), mvpci_drv->mmio + (adr));                             \
    }
#define mvwrite64(dat, adr)                                                 \
    {                                                                       \
        dprintk(0,                                                          \
            "mvpci-%d: writing 0x%08x to register @ 0x%08x (0x%08lx)...\n", \
            mvpci_drv->nr, (dat), (adr), mvpci_drv->mmio + (adr));          \
        writeq((dat), mvpci_drv->mmio + (adr));                             \
    }
#else
#define mvwrite(dat, adr) writel((dat), mvpci_drv->mmio + (adr))
#define mvwrite64(dat, adr) writeq((dat), mvpci_drv->mmio + (adr))
#endif
#define mvread(adr) readl(mvpci_drv->mmio + (adr))
#define mvread64(adr) readq(mvpci_drv->mmio + (adr))

struct mvpci_props {
    u32 debug, version;
};

#define MVPCI_MSI_MIN_VECTORS 1 /* max. = 32, power of 2 */
#define MVPCI_MSI_MAX_VECTORS 4 /* max. = 32, power of 2 */
#define MVPCI_MAX_UIO_VECTORS MVPCI_MSI_MAX_VECTORS
#define MVPCI_REMEMBERED_IRQS 0x0820    /* read to get type or interrupt */
#define MVPCI_IO_CONTROL      0x60000   /* used to enable and disable IO interrupts */
#define IRQ_SOURCE_MSEQ       0x1   /* interrupt source microsequencer */
#define IRQ_SOURCE_1          0x2   /* interrupt source 1 (I/O) */
#define IRQ_SOURCE_2          0x4   /* interrupt source 2 (statistics) */
#define IRQ_SOURCE_3          0x8   /* interrupt source 3 (sc) */

//#define USE_IOW_SEQUENCER     /* memory for the sequencer not yet used */

struct mvpci_driver;            /* forward declaration */

struct mvpci_fh {
    struct mvpci_driver *mvpci_drv;
    struct file *owner;
    spinlock_t dma_queue_lock;
    unsigned opened;
};

// forward declaration
struct mvdma_device;

struct mvpci_driver {
    struct pci_dev *pci;
#ifdef CONFIG_NUMA
    int node;
#endif                          /* CONFIG_NUMA */
    struct cdev cdev;
    struct mvpci_fh fh;
    struct mvdma_device *pdma;
    struct mutex dev_lock;
    int nr;
    unsigned users;
    char name[16];
    unsigned short id;
    uint8_t revision;
    uint16_t subsystem_id;
    uint32_t flags;
    void __iomem *mmio;
    int initialized;
    struct msix_entry msi_x_entry[MVPCI_MSI_MAX_VECTORS];
    struct uio_info uio_info[MVPCI_MAX_UIO_VECTORS];
    struct class_device *class_device;
    dev_t devnum;
    uint32_t maxpayloadsize;
    uint32_t maxreadreq;
    struct mvpci_props props;
#ifdef USE_IOW_SEQUENCER
    void *iow_sequencer;
    dma_addr_t dma_iow_sequencer;
#endif // USE_IOW_SEQUENCER
    int coherent;
    uint32_t saved_mps;
    u64 read_timeout_count;
    struct hrtimer read_hrtimer;
    uint32_t streaming_owner;
    uint32_t user_count;
};

struct mvpci_sysfs {
    struct kobj_attribute camera_reset_attribute;
    struct attribute *camera_attrs[2];
    struct attribute_group camera_attr_group;
};

extern u32 debug;
extern u32 isr_debug;

#endif /* __KERNEL__ */
#endif /* __MVPCI_H__ */
