/*
 * mvPCIe DMA Driver for BVS CA-BN (mvBlueNAOS) series cameras
 *
 * SPDX-FileCopyrightText: 2019 - 2023 MATRIX VISION GmbH <info@matrix-vision.de>
 * SPDX-FileCopyrightText: 2023 Balluff GmbH
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */

#include "linux/bitmap.h"
#include "mvdma.h"

/* ******************************************************************************** */
typedef union {
    struct {
        uint32_t p_num:5;       /*  5 bits for the page number 0 .. 31 */
        uint32_t b_num:11;      /* 11 bits for the block number 0 .. 2047 */
        uint32_t pages:5;       /*  5 bits for the size 0 .. 31 */
        uint32_t reserved:11;   /* 11 bits reserved / unused */
    } u;
    uint32_t h;                 /* handle as a 32 bit integer */
} plhandle;

/* *********************************************************************************************** */
static int construct_id(struct mvdma_device *pdma, uint16_t block_num, uint16_t page_num, uint16_t num_pages,
                        plhandle * id)
{
    if ((page_num >= PAGES_PER_BLOCK) || (block_num >= MAX_POOL_BLOCKS) || (num_pages > PAGES_PER_BLOCK)) {
        return -1;
    }
    id->u.p_num = page_num;
    id->u.b_num = block_num;
    id->u.pages = num_pages - 1;
    //dprintk_dma(MVDMA_DEBUG_POOL, "HANDLE: 0x%08x, b_num = 0x%x, p_num = 0x%x, pages = 0x%x\n", id->h, id->u.b_num, id->u.p_num, id->u.pages);
    return 0;
}

/* *********************************************************************************************** */
int reserve_page_list(struct mvdma_device *pdma, struct page_list_req *plr)
{
    unsigned long search_from = 0;

    /* check that we have at least one page free */
    if (unlikely(bitmap_full(pdma->page_list_pool_map, PAGES_IN_POOL))) {
        pr_err("%s: ERROR! There is no space left in the page list pool!\n", __FUNCTION__);
        return -ENOMEM;
    }

    do {
        unsigned long first_zero =
            bitmap_find_next_zero_area(pdma->page_list_pool_map, PAGES_IN_POOL, search_from, plr->num_pages,
                                       0 /* = alignment */ );
        if (first_zero == PAGES_IN_POOL) {
            dprintk_dma(MVDMA_DEBUG_POOL, "ERROR!: cannot find 0x%x pages free in the bitmap!\n", plr->num_pages);
            return -ENOMEM;
        } else {                /* found a free area big enough */

            unsigned long last_zero = first_zero + plr->num_pages - 1;
            uint16_t block_num_start = (uint16_t) (first_zero / PAGES_PER_BLOCK);
            uint16_t block_num_end = (uint16_t) (last_zero / PAGES_PER_BLOCK);

            if (block_num_start != block_num_end) {
                dprintk_dma(MVDMA_DEBUG_POOL, "keep searching starting at the next block (%d)....\n", block_num_end);
                search_from = block_num_end * PAGES_PER_BLOCK;
            } else {
                uint16_t page_num = (uint16_t) (first_zero % PAGES_PER_BLOCK);
                plhandle *id = (plhandle *) (&plr->handle);
                const uint64_t offset = page_num * (PAGE_SIZE / sizeof(uint64_t));

                dprintk_dma(MVDMA_DEBUG_POOL, "set bits in the range 0x%lx - 0x%lx within block 0x%x\n", first_zero,
                            last_zero, block_num_start);
                bitmap_set(pdma->page_list_pool_map, first_zero, plr->num_pages);

                dprintk_dma(MVDMA_DEBUG_POOL,
                            "Constructing handle for block 0x%04x and page 0x%02x with 0x%02x pages\n", block_num_start,
                            page_num, plr->num_pages);
                construct_id(pdma, block_num_start, page_num, plr->num_pages, id);

                plr->addr = (uint64_t *) pdma->page_list[block_num_start] + offset;
#ifdef USE_COHERENT_FOR_PAGE_LISTS
                if (pdma->mvpci_dev->coherent) {
                    plr->dma_addr = pdma->page_list_dma[block_num_start] + (sizeof(uint64_t) * offset);
                } else
#endif // USE_COHERENT_FOR_PAGE_LISTS
                {
                    /* convert the address of the 1st. entry in the page list to a DMA address and store it */
                    /* this is always written by the CPU and read by the FPGA, so use DMA_TO_DEVICE */
                    plr->dma_addr = dma_map_single(pdma->dev, plr->addr, PAGE_SIZE * plr->num_pages, DMA_TO_DEVICE);
                    if (dma_mapping_error(pdma->dev, plr->dma_addr)) {
                        dprintk_dma(MVDMA_DEBUG_POOL, "ERROR!: cannot perform dma_map_single!\n");
                        return -ENOMEM;
                    }
                }
                dprintk_dma(MVDMA_DEBUG_POOL, "plr->addr = 0x%px, plr->dma_addr = 0x%llx, offset = 0x%08llx\n", plr->addr, plr->dma_addr, offset);

                /* normal return here */
                return 0;
            }
        }
    } while (search_from < PAGES_IN_POOL);

    pr_err("ERROR!: cannot find 0x%x pages free in the bitmap!\n", plr->num_pages);
    return -ENOMEM;
}

/* *********************************************************************************************** */
int free_page_list(struct mvdma_device *pdma, struct page_list_req *plr)
{
    plhandle *ph = (plhandle *) & plr->handle;
    unsigned long b;

    if (plr->addr && plr->dma_addr) {
        unsigned long bit = (ph->u.b_num * PAGES_PER_BLOCK) + ph->u.p_num;

        dprintk_dma(MVDMA_DEBUG_POOL, "clear bits in the range 0x%lx - 0x%lx within block 0x%x\n", bit,
                    bit + ph->u.pages, ph->u.b_num);

        for (b = bit; b <= (bit + ph->u.pages); b++) {
            if (test_bit(b, pdma->page_list_pool_map)) {
                clear_bit(b, pdma->page_list_pool_map);
            } else {
                dprintk_dma(MVDMA_DEBUG_POOL, "ERROR! bit 0x%lx was not set!\n", bit);
                return -1;
            }
        }
#ifdef USE_COHERENT_FOR_PAGE_LISTS
        if (!pdma->mvpci_dev->coherent)
#endif // USE_COHERENT_FOR_PAGE_LISTS
        { /* unmap the DMA address of the entry in the page list */
            dprintk_dma(MVDMA_DEBUG_POOL, "dma_unmap 0x%08llx\n", plr->dma_addr);
            dma_unmap_single(pdma->dev, plr->dma_addr, PAGE_SIZE * plr->num_pages, DMA_TO_DEVICE);
        }
        /* clear the plr completely */
        memset(plr, 0, sizeof(struct page_list_req));
    }
    return 0;
}

/* *********************************************************************************************** */
int init_page_list_pool(struct mvdma_device *pdma)
{
    int i;

#ifdef USE_COHERENT_FOR_PAGE_LISTS
    if (pdma->mvpci_dev->coherent) {
        printk("Assuming coherent memory available for page lists.\n" );
    } else
#endif // USE_COHERENT_FOR_PAGE_LISTS
    {
        printk("NOT using coherent memory for page lists.\n" );
    }

    for (i = 0; i < MVDMA_PAGE_LIST_CNT; i++) {
#ifdef USE_COHERENT_FOR_PAGE_LISTS
        if (pdma->mvpci_dev->coherent) {
            pdma->page_list[i] = (struct mvdma_page_list *)dma_alloc_coherent(pdma->dev, sizeof(struct mvdma_page_list), &pdma->page_list_dma[i], GFP_KERNEL);
            dprintk_dma(MVDMA_DEBUG_POOL, "%s:  page_list[%d] = 0x%px, dma_addr = 0x%08llx\n", __FUNCTION__, i, pdma->page_list[i], pdma->page_list_dma[i]);
        } else
#endif // USE_COHERENT_FOR_PAGE_LISTS
        {
            pdma->page_list[i] = (struct mvdma_page_list *)kzalloc(sizeof(struct mvdma_page_list), GFP_KERNEL);
            dprintk_dma(MVDMA_DEBUG_POOL, "%s:  page_list[%d] = 0x%px\n", __FUNCTION__, i, pdma->page_list[i]);
        }
        if (!pdma->page_list[i]) {
            pr_err("%s: ERROR! kzalloc for mvdma_page_list %d failed\n", __FUNCTION__, i);
            return -ENOMEM;
        }
    }
    return 0;
}

/* *********************************************************************************************** */
int clear_page_list_pool(struct mvdma_device *pdma)
{
    if (!bitmap_empty(pdma->page_list_pool_map, MVDMA_PAGE_LIST_CNT)) {
        dprintk_dma(MVDMA_DEBUG_POOL, "%s: WARNING! page list bool BITMAP was not empty!\n", __FUNCTION__);
    }
    bitmap_zero(pdma->page_list_pool_map, MVDMA_PAGE_LIST_CNT);
    dprintk_dma(MVDMA_DEBUG_POOL, "%s: clear pool mapping\n", __FUNCTION__);
    return 0;
}

/* *********************************************************************************************** */
int free_page_list_pool(struct mvdma_device *pdma)
{
    int i;

    clear_page_list_pool(pdma);
    dprintk_dma(MVDMA_DEBUG_POOL, "freeing page_list\n");
    for (i = 0; i < MVDMA_PAGE_LIST_CNT; i++) {
        if (pdma->page_list[i]) {
#ifdef USE_COHERENT_FOR_PAGE_LISTS
            if (pdma->mvpci_dev->coherent) {
                //dprintk_dma(MVDMA_DEBUG_POOL, "freeing coherent page_list[%d]\n", i);
                dma_free_coherent(pdma->dev, sizeof(struct mvdma_page_list), pdma->page_list[i], pdma->page_list_dma[i]);
            } else
#endif // USE_COHERENT_FOR_PAGE_LISTS
            {
                //dprintk_dma(MVDMA_DEBUG_POOL, "freeing kfree page_list[%d]\n", i);
                kfree(pdma->page_list[i]);
            }
        }
    }
    return 0;
}
